from typing import Iterator, Generic, TypeVar
from abc import ABC
from typing import Type
from localis.data import DTO, Model
from localis.search import FuzzySearch

TModel = TypeVar("TModel", bound=Model)
TDTO = TypeVar("TDTO", bound=DTO)


class Registry(Generic[TModel, TDTO], ABC):
    """Abstract base registry class defining interface for lookup and search."""

    ID_FIELDS: tuple[str] = ()

    SEARCH_FIELD_WEIGHTS: dict[str, float] = {}
    SEARCH_ORDER_FIELDS: list[str] = []
    """Override to provide the fields for search scoring."""

    def __init__(self, model_cls: Type[TModel]):
        self._model_cls: Type[TModel] = model_cls
        self._count: int | None = None
        self._cache: list[TDTO] = None
        self._order_by: str = ""
        self._addl_search_attrs: list[str] = []

    @property
    def cache(self):
        if self._cache is None:
            self._cache = [m.to_dto() for m in self._model_cls.select()]
        return self._cache

    def __iter__(self) -> Iterator[TDTO]:
        return iter(self.cache)

    def __getitem__(self, index: int | slice) -> TDTO | list[TDTO]:
        return self.cache[index]

    def __len__(self) -> int:
        if self._count is None:
            self._count = self._model_cls.count()
        return self._count

    @property
    def count(self) -> int:
        return self.__len__()

    def get(self, *, id: int | None = None, **kwargs) -> TDTO | None:
        return None

    def filter(
        self, query: str = None, name: str = None, limit: int = None, **kwargs
    ) -> list[TDTO]:
        if kwargs:
            if name:
                kwargs.update({"name": name})
            results = self._model_cls.fts_match(
                field_queries=kwargs, order_by=["rank"], limit=limit
            )

        elif name:
            results = self._model_cls.fts_match(
                field_queries={"name": name}, order_by=["rank"], limit=limit
            )
        elif query:
            results = self._model_cls.fts_match(query, order_by=["rank"], limit=limit)
        else:
            return []

        return [r.to_dto() for r in results]

    def search(self, query: str, limit=None, **kwargs) -> list[tuple[TDTO, float]]:
        if not query:
            return []

        search = FuzzySearch(
            query,
            self._model_cls,
            self.SEARCH_FIELD_WEIGHTS,
            self.SEARCH_ORDER_FIELDS,
            limit,
        )

        return search.run()

    def _sort_matches(self, matches: list, limit: int) -> list[TDTO]:
        return [
            (row_data.dto, score)
            for row_data, score in sorted(matches, key=lambda r: r[1], reverse=True)[
                :limit
            ]
        ]



def prep_tokens(query: str, exact_match: bool) -> str:
    """Wrap each token in quotes and at * wildcard for prefixing if not exact_match"""

    tokens = query.split()
    for i, token in enumerate(tokens):
        tokens[i] = f'"{token}"'
        if not exact_match:
            tokens[i] += "*"
    return " ".join(tokens)


def clean_row(row: dict[str, str]) -> dict[str, str | None]:
    """Clean and normalize all fields in a dict"""
    return {k: (v if v.strip() != "" else None) for k, v in row.items()}


def chunked(list: list, size: int):
    for i in range(0, len(list), size):
        yield list[i : i + size]



from localis.data.models.model import Model
from localis.data.models.fields import CharField, IntField, FloatField
from localis.dtos import SubdivisionBasic, City
from localis.utils import clean_row, chunked
import csv


class CityModel(Model[City]):
    table_name = "cities"
    dto_class = City

    name = CharField()
    geonames_id = CharField()
    admin1 = CharField()
    admin2 = CharField()
    country = CharField()
    alt_names = CharField()
    population = IntField(index=False)
    lat = FloatField(index=False)
    lng = FloatField(index=False)

    def to_dto(self) -> City:

        def parse_subdivision(raw_sub: str | None, lvl: int) -> SubdivisionBasic | None:
            if raw_sub:
                name, geonames_code, iso_code = raw_sub.split("|")
                return SubdivisionBasic(name, geonames_code, iso_code, lvl)

        # build subdivision DTOs
        admin1 = parse_subdivision(self.admin1, 1)
        admin2 = parse_subdivision(self.admin2, 2)
        subdivisions = [s for s in (admin1, admin2) if s is not None]

        # parse country
        country_parts = self.country.split("|")
        country = country_parts[0] if country_parts else None
        alpha2 = country_parts[1] if len(country_parts) > 1 else None
        alpha3 = country_parts[2] if len(country_parts) > 2 else None

        # build display name
        display_parts = [self.name, *[s.name for s in subdivisions[::-1]], country]
        display_name = ", ".join(display_parts)

        return City(
            id=self.id,
            geonames_id=int(self.geonames_id),
            name=self.name,
            display_name=display_name,
            subdivisions=subdivisions,
            country=country,
            country_alpha2=alpha2,
            country_alpha3=alpha3,
            alt_names=self.alt_names.split("|") if self.alt_names else [],
            population=int(self.population),
            lat=float(self.lat),
            lng=float(self.lng),
        )

    @classmethod
    def load(cls, file):
        cls.db.create_tables([CityModel])
        cities: list[dict] = []
        reader = csv.DictReader(file, delimiter="\t")

        for row in reader:
            MAX_DIGITS = 9
            row["population"] = f"{int(row['population']):0{MAX_DIGITS}d}"
            cities.append(clean_row(row))

        with cls.db.atomic():
            for batch in chunked(list(cities), 1000):
                try:
                    cls.insert_many(batch)
                except Exception as e:
                    print(f"Unexpected error on batch: {e}")
                    raise e

    def __init__(
        self,
        geonames_id: int,
        name: str,
        admin1: str,
        admin2: str,
        country: str,
        alt_names: str,
        population: int,
        lat: float,
        lng: float,
        **kwargs,
    ):
        self.geonames_id = geonames_id
        self.name = name
        self.admin1 = admin1 or ""
        self.admin2 = admin2 or ""
        self.country = country or ""
        self.alt_names = alt_names or ""
        self.population = population
        self.lat = lat
        self.lng = lng

        super().__init__(**kwargs)



from localis.data.models.model import Model
from localis.data.models.fields import CharField, IntField
from localis.dtos import Country


class CountryModel(Model[Country]):
    table_name = "countries"
    dto_class = Country

    name = CharField()
    official_name = CharField()
    alpha2 = CharField()
    alpha3 = CharField()
    numeric = IntField()
    alt_names = CharField()
    flag = CharField(index=False)

    def to_dto(self):
        self.numeric = int(self.numeric) if self.numeric else None
        self.alt_names = self.alt_names.split("|") if self.alt_names else []
        return super().to_dto()

    def __init__(
        self,
        name: str,
        official_name: str,
        alpha2: str,
        alpha3: str,
        numeric: int,
        alt_names: str,
        flag: str,
        **kwargs
    ):
        self.name = name
        self.official_name = official_name or ""
        self.alpha2 = alpha2
        self.alpha3 = alpha3
        self.numeric = numeric
        self.alt_names = alt_names or ""
        self.flag = flag

        super().__init__(**kwargs)



from localis.data import db, Database


class MetaStore:
    """A simple model and cache for a meta table for storing flags"""

    db: Database = db

    def __init__(self):
        self._cache = {}

    @classmethod
    def create_table(cls):
        cls.db.execute(
            """
                CREATE TABLE IF NOT EXISTS meta(
                    key TEXT PRIMARY KEY,
                    value TEXT)"""
        )

        cls.db.execute("UPDATE meta SET value = 0 WHERE value = 1")
        cls.db.commit()

    def get(self, key: str):

        if key in self._cache:
            return self._cache[key]

        row = self.db.execute("SELECT value FROM meta WHERE key = ?", (key,)).fetchone()

        value = row["value"] if row else None
        self._cache[key] = value
        return value

    def set(self, key: str, value: str):
        self.db.execute(
            "INSERT OR REPLACE INTO meta (key, value) VALUES (?, ?)", (key, value)
        )
        self.db.commit()
        self._cache[key] = value



from localis.registries.registry import Registry
from localis.data import CityModel, City, MetaStore, db
import requests
import io
import localis
import typer
import os


class CityRegistry(Registry[CityModel, City]):
    """Registry for cities"""

    ID_FIELDS = ("id", "geonames_id")

    SEARCH_FIELD_WEIGHTS = {
        "name": 1.0,
        "alt_names": 1.5,
        "admin1": 0.2,
        "admin2": 0.1,
        "country": 0.2,
    }

    SEARCH_ORDER_FIELDS = ["population"]

    META_URL_KEY = "cities_tsv_url"

    def __init__(self, model_cls):
        super().__init__(model_cls)
        self._order_by = "population DESC"

        self._meta = MetaStore()
        self._loaded = False
        """WARNING: Do not mutate directly, controlled by set_loaded()"""
        self.set_loaded()

    def set_loaded(self) -> bool:
        self._loaded = db.CONFIG_FILE.exists()

    def load(self, confirmed: bool = False, custom_dir: str = "") -> None:
        if self._loaded:
            print("Cities data already loaded.")
            return

        # USER CONFIRMATION
        confirmed = confirmed or typer.confirm(
            "Loading cities is HEAVY, there are nearly half a million entries and it expands the database to 200MB+. Proceeding with load will copy the sqlite database to your project root, download cities.tsv and load it into the copied database and update your .gitignore. Are you sure you want to proceed?"
        )

        # DID NOT CONFIRM
        if not confirmed:
            typer.echo("Aborting load.")
            raise typer.Exit()

        # COPY DATABASE
        typer.echo(f"Copying database to {custom_dir}/{db.FILENAME}...")
        path = db.copy_to(custom_dir)
        db.set_db_path(path)

        # UPDATE GITIGNORE
        typer.echo(f"Updating .gitignore...")

        gitignore_path = ".gitignore"

        if os.path.exists(gitignore_path):
            with open(gitignore_path, "r") as f:
                existing_content = f.read()
                if db.FILENAME in existing_content:
                    typer.echo("Database already in .gitignore. Skipping...")
        else:
            existing_content = ""

        with open(gitignore_path, "a") as f:
            if existing_content and not existing_content.endswith("\n"):
                f.write("\n")
            f.write(db.FILENAME)

        # DOWNLOAD TSV FIXTURE
        url = self._meta.get(self.META_URL_KEY)
        if url:
            typer.echo("Downloading TSV fixture...")

            try:
                response = requests.get(url)
                response.raise_for_status()  # just to be safe
            except requests.HTTPError as e:
                if e.response.status_code == 404:
                    e.add_note(
                        f"There is a problem with the cities.tsv url, please raise a new issue: https://github.com/dstoffels/localis/issues.\nurl: {url}"
                    )
                raise e

            # LOAD TSV INTO DATABASE
            typer.echo("TSV fixture downloaded, loading cities into database...")
            tsv = io.StringIO(response.text)
            CityModel.load(tsv)

            self.set_loaded()
            typer.echo(f"{self.count} cities loaded.")
            typer.echo(
                "Run 'localis unload cities' in the CLI or 'localis.cities.unload()' to revert."
            )

        else:
            raise ValueError(
                f"Error fetching the cities fixture url, the database (meta table) may have been corrupted. Please submit a new issue: https://github.com/dstoffels/localis/issues.\nCurrent url: {url}"
            )

    def unload(self) -> None:
        if not self._loaded:
            typer.echo("No cities to unload.")
        else:
            # UNLOAD FILES
            typer.echo(f"Removing database and localis.conf...")
            db.revert_to_default()
            typer.echo("Files removed.")

            # UPDATE GITIGNORE
            typer.echo("Updating .gitignore...")
            with open(".gitignore", "r+") as f:
                content = f.read()
                content = content.replace(db.FILENAME, "")
                f.write(content)

            self.set_loaded()
            self._count = None
            typer.echo("Cities successfully unloaded from db.")

    @property
    def count(self):
        self._check_loaded()
        return super().count

    def _check_loaded(self):
        if not self._loaded:
            raise RuntimeError(
                "Cities data not yet loaded. Load with `localis.cities.load()` or `localis load cities` from the CLI."
            )

    def get(self, *, id: int = None, geonames_id: str = None, **kwargs):
        self._check_loaded()
        cls = self._model_cls

        field_map = {
            "id": None,
            "geonames_id": cls.geonames_id,
        }

        model = None
        for arg, field in field_map.items():
            val = locals()[arg]
            if val is not None:
                model = cls.get_by_id(val) if arg == "id" else cls.get(field == val)

        return model.to_dto() if model is not None else None

    def filter(
        self,
        query: str = None,
        name: str = None,
        limit: int = None,
        admin1: str = None,
        admin2: str = None,
        country: str = None,
        alt_name: str = None,
        **kwargs,
    ):
        self._check_loaded()
        if kwargs:
            return []
        if admin1:
            kwargs["admin1"] = admin1
        if admin2:
            kwargs["admin2"] = admin2
        if country:
            kwargs["country"] = country
        if alt_name:
            kwargs["alt_names"] = alt_name

        return super().filter(query, name, limit, **kwargs)

    def search(self, query, limit=None, **kwargs):
        self._check_loaded()
        return super().search(query, limit, **kwargs)

    def for_country(
        self,
        *,
        id: int = None,
        alpha2: str = None,
        alpha3: str = None,
        numeric: int = None,
        population__lt: int | None = None,
        population__gt: int | None = None,
        **kwargs,
    ) -> list[City]:
        self._check_loaded()
        if population__gt is not None and population__lt is not None:
            raise ValueError("population__gt and population__lt are mutually exclusive")

        provided = {
            k: v
            for k, v in locals().items()
            if k in localis.countries.ID_FIELDS and v is not None
        }
        country = localis.countries.get(**provided)
        if country is None:
            return []

        country_field = "|".join([country.name, country.alpha2, country.alpha3])
        results: list[CityModel] = self._model_cls.select(
            CityModel.country == country_field
        )

        dtos = [r.to_dto() for r in results]
        if population__gt is not None:
            return [d for d in dtos if d.population > population__gt]
        elif population__lt is not None:
            return [d for d in dtos if d.population < population__lt]
        else:
            return dtos

    def for_subdivision(
        self,
        *,
        id: int = None,
        geonames_code: str = None,
        iso_code: str = None,
        population__lt: int | None = None,
        population__gt: int | None = None,
        **kwargs,
    ) -> list[City]:
        self._check_loaded()
        if population__gt is not None and population__lt is not None:
            raise ValueError("population__gt and population__lt are mutually exclusive")

        provided = {
            k: v
            for k, v in locals().items()
            if k in localis.subdivisions.ID_FIELDS and v is not None
        }

        sub = localis.subdivisions.get(**provided)

        if sub is None:
            return []

        sub_field = "|".join([sub.name, sub.geonames_code, sub.iso_code])
        results: list[CityModel] = self._model_cls.select(CityModel.admin1 == sub_field)

        if not results:
            results = self._model_cls.select(CityModel.admin2 == sub_field)

        dtos = [r.to_dto() for r in results]
        if population__gt is not None:
            return [d for d in dtos if d.population > population__gt]
        elif population__lt is not None:
            return [d for d in dtos if d.population < population__lt]
        else:
            return dtos



from localis.registries import CountryRegistry, SubdivisionRegistry, CityRegistry
from localis.data import CountryModel, SubdivisionModel, CityModel
from localis.dtos import Country, Subdivision, City

countries = CountryRegistry(CountryModel)
"""The countries registry

- Registries are iterable. NOTE: caches all entries on first iterable access. 
- To manually trigger caching, access `localis.countries.cache` (property).
"""

subdivisions = SubdivisionRegistry(SubdivisionModel)
"""The subdivisions registry

- Registries are iterable. NOTE: caches all entries (51k entries, 30MB) on first iterable access. 
- To manually trigger caching, access `localis.subdivisions.cache` (property).
"""

cities = CityRegistry(CityModel)
"""The cities registry.

- Must be loaded before use: `localis.cities.load()` or via CLI `localis load cities`.
- WARNING: Loading this registry expands the database to ~250MB.
- Registries are iterable. WARNING: caches all entries (451k cities, 461MB) on first iterable access.
- To manually trigger caching, access `localis.cities.cache` (property).
"""



from localis.data.models.model import Model
from localis.data.models.fields import CharField
from localis.dtos import Subdivision


class SubdivisionModel(Model[Subdivision]):
    table_name = "subdivisions"
    dto_class = Subdivision

    name = CharField()
    alt_names = CharField()
    type = CharField()
    geonames_code = CharField()
    iso_code = CharField()
    country = CharField()
    parent_id = CharField()

    def to_dto(self):
        self.alt_names = self.alt_names.split("|") if self.alt_names else []
        self.parent_id = int(self.parent_id) if self.parent_id else None
        self.admin_level = 2 if self.parent_id else 1
        self.country, self.country_alpha2, self.country_alpha3 = self.country.split("|")
        return super().to_dto()

    def __init__(
        self,
        name: str,
        alt_names: str,
        type: str,
        geonames_code: str,
        iso_code: str,
        country: str,
        parent_id: str,
        **kwargs
    ):
        self.name = name
        self.alt_names = alt_names or ""
        self.type = type
        self.geonames_code = geonames_code
        self.iso_code = iso_code
        self.country = country
        self.parent_id = parent_id

        super().__init__(**kwargs)



import typer
from typing_extensions import Annotated
from localis.data import MetaStore, Database
from localis import CityRegistry
import localis


app = typer.Typer(help="localis CLI")
meta = MetaStore()


@app.command()
def load(
    dataset: str,
    confirmed: Annotated[
        bool, typer.Option("--yes", "-y", help="Auto-confirm without prompting")
    ] = False,
    custom_dir: Annotated[
        str,
        typer.Option(
            "--path",
            "-p",
            help="Custom directory to store the database, by default load will copy the sqlite file to your current working directory.",
        ),
    ] = "",
):
    """Load a remote dataset into the database. Use `localis status` to view eligible datasets."""

    DATASETS = {"cities": lambda: localis.cities.load(confirmed, custom_dir)}

    if dataset in DATASETS:
        DATASETS[dataset]()
    else:
        typer.echo(f"Unknown dataset.")
        raise typer.Exit(code=1)


@app.command()
def unload(dataset: str):
    """Drop an eligible table from the database. Use `localis status` to view eligible datasets."""
    DATASETS = {"cities": localis.cities.unload}

    if dataset in DATASETS:
        DATASETS[dataset]()
    else:
        typer.echo(f"Unknown dataset.")
        raise typer.Exit(code=1)


@app.command(hidden=True)
def seturl(url: str):
    meta.set(CityRegistry.META_URL_KEY, url)
    typer.echo(
        f"Fixture url set to: {meta.get(CityRegistry.META_URL_KEY)}",
    )


@app.command()
def status():
    """View the current state of datasets"""
    url_key = CityRegistry.META_URL_KEY
    typer.echo(f"{url_key}: {meta.get(url_key)}")
    typer.echo(f"Current Daabase: {Database.get_db_path()}")
    typer.echo(f"Cities Loaded: {Database.CONFIG_FILE.exists()}")



from localis.registries.registry import Registry
from localis.dtos import Subdivision
from localis.data import SubdivisionModel, CountryModel
from rapidfuzz import fuzz
import localis


class SubdivisionRegistry(Registry[SubdivisionModel, Subdivision]):
    """
    Registry for managing Subdivision entities.

    Supports exact lookup by ISO code, alpha2, or country code,
    fuzzy search by these keys, and filtering by country or country code.
    """

    ID_FIELDS = ("id", "geonames_code", "iso_code")

    SEARCH_FIELD_WEIGHTS = {"name": 1.0, "alt_names": 0.4, "country": 0.33}

    def get(
        self,
        *,
        id: int = None,
        iso_code: str = None,
        geonames_code: str = None,
        **kwargs
    ):
        cls = self._model_cls

        field_map = {
            "id": None,
            "iso_code": cls.iso_code,
            "geonames_code": cls.geonames_code,
        }

        model = None
        for arg, field in field_map.items():
            val = locals()[arg]
            if val is not None:
                model = cls.get_by_id(val) if arg == "id" else cls.get(field == val)

        return model.to_dto() if model is not None else None

    def filter(
        self,
        query: str = None,
        name: str = None,
        limit: int = None,
        type: str = None,
        country: str = None,
        alt_name: str = None,
        **kwargs
    ):
        if kwargs:
            return []
        if type is not None:
            kwargs["type"] = type
        if country is not None:
            kwargs["country"] = country
        if alt_name is not None:
            kwargs["alt_names"] = alt_name

        return super().filter(query, name, limit, **kwargs)

    def for_country(
        self,
        *,
        admin_level: int | None = 1,
        id: int = None,
        alpha2: str = None,
        alpha3: str = None,
        numeric: int = None,
        **kwargs
    ) -> list[Subdivision]:
        """Get all subdivisions for a given country by id, alpha2, alpha3 or numeric code. Can filter results by admin_level (default=1)."""
        provided = {
            k: v
            for k, v in locals().items()
            if k in ("id", "alpha2", "alpha3", "numeric") and v is not None
        }
        country = localis.countries.get(**provided)
        if country is None:
            return []

        country_field = "|".join([country.name, country.alpha2, country.alpha3])
        results: list[SubdivisionModel] = self._model_cls.select(
            SubdivisionModel.country == country_field
        )

        dtos = [r.to_dto() for r in results]

        return [d for d in dtos if d.admin_level == admin_level or admin_level == None]

    def types_for_country(
        self,
        *,
        admin_level: int | None = 1,
        id: int = None,
        alpha2: str = None,
        alpha3: str = None,
        numeric: int = None,
        **kwargs
    ) -> list[str]:
        """Fetch a list of distinct subdivision types for a given country by id, alpha2, alpha3 or numeric code. Can filter results by admin level (default=1)"""
        provided = {
            k: v
            for k, v in locals().items()
            if k in localis.countries.ID_FIELDS and v is not None
        }
        results = self.for_country(admin_level=admin_level, **provided)

        types = set(
            [
                r.type
                for r in results
                if r.type is not None and r.admin_level == admin_level
            ]
        )

        return list(types)



from abc import ABC
from typing import Literal
from typing import Generic, TypeVar

T = TypeVar("T")


class Expression:
    def __init__(self, sql: str, params: tuple = ()):
        self.sql = sql
        self.params = params

    def __and__(self, other):
        return Expression(f"({self.sql} AND {other.sql})", self.params + other.params)

    def __or__(self, other):
        return Expression(f"({self.sql} OR {other.sql})", self.params + other.params)

    def __str__(self):
        return self.sql


class Field(Generic[T], ABC):
    type: Literal["TEXT", "INTEGER", "FLOAT", "BOOLEAN"] = "INTEGER"

    def __init__(
        self,
        *,
        nullable: bool = True,
        default: T | None = None,
        unique: bool = False,
        index: bool = True,
        primary_key: bool = False,
    ):
        self.nullable = nullable
        self.default = default
        self.unique = unique
        self.index = index
        self.primary_key = primary_key
        self.name = None  # set by the model

    def __set_name__(self, owner, name):
        self.name = name
        self.model = owner

    def __get__(self, instance, owner) -> T:
        if instance is None:
            return self
        return instance.__dict__.get(self.name, self.default)

    def __set__(self, instance, value: T):
        instance.__dict__[self.name] = value

    def __eq__(self, other):
        return Expression(f"{self.name} = ?", (str(other),))

    def __ne__(self, other):
        return Expression(f"{self.name} != ?", (str(other),))

    def __lt__(self, other):
        return Expression(f"{self.name} < ?", (str(other),))

    def __le__(self, other):
        return Expression(f"{self.name} <= ?", (str(other),))

    def __gt__(self, other):
        return Expression(f"{self.name} > ?", (str(other),))

    def __ge__(self, other):
        return Expression(f"{self.name} >= ?", (str(other),))

    def like(self, pattern):
        return Expression(f"{self.name} LIKE ?", (pattern,))

    def isin(self, values: list):
        placeholders = ", ".join("?" for _ in values)
        return Expression(f"{self.name} IN ({placeholders})", tuple(values))

    def get_sql(self) -> str:
        parts = [self.name, self.type]
        if self.primary_key:
            parts.append("PRIMARY KEY")
        if not self.nullable:
            parts.append("NOT NULL")
        if self.unique:
            parts.append("UNIQUE")
        if self.default is not None:
            val = self.default
            if isinstance(self.default, str):
                val = f"'{val}'"
            elif isinstance(self.default, bool):
                val = int(val)
            parts.append(f"DEFAULT {self.default}")
        parts.append("COLLATE NOCASE")
        return " ".join(parts)

    def get_idx(self, name: str, table: str) -> str:
        if self.index:
            return f"CREATE INDEX IF NOT EXISTS {table}_{name}_idx ON {table}({name})"
        return ""


class AutoField(Field[int]):
    def get_sql(self):
        return f"{self.name} INTEGER PRIMARY KEY AUTOINCREMENT"


class CharField(Field[str]):
    type = "TEXT"


class IntField(Field[int]):
    pass


class FloatField(Field[float]):
    type = "FLOAT"


class BooleanField(Field[bool]):
    type = "BOOLEAN"


class ForeignKeyField(Field[int]):
    def __init__(
        self,
        *args,
        references: str = "",
        field: str = "id",
        on_delete: Literal[
            "CASCADE", "RESTRICT", "SET NULL", "SET DEFAULT"
        ] = "SET NULL",
        nullable: bool = None,
        **kwargs,
    ):
        if nullable is not None:
            kwargs["nullable"] = nullable
        else:
            kwargs["nullable"] = False
        super().__init__(*args, **kwargs)
        self.references = references
        self.field = field
        self.on_delete = on_delete

    def get_sql(self):
        parts = [
            self.name,
            self.type,
            f"REFERENCES {self.references}({self.field}) ON DELETE {self.on_delete}",
        ]
        if not self.nullable:
            parts.insert(2, "NOT NULL")
        return " ".join(parts)



import atexit
from typing import Any, Dict, List, Tuple
from contextlib import contextmanager
import sqlite3
from importlib import resources
import shutil
from pathlib import Path


class Database:
    MAX_PREFIX = 24
    PATH = "localis.data"
    FILENAME = "localis.db"
    CONFIG_FILE = Path.cwd() / ".localis.conf"

    def __init__(self, db_path: str = None):
        self.db_path: str = db_path or self.get_db_path()
        self._conn: sqlite3.Connection | None = None
        self._setup_conn()

    def _setup_conn(self) -> None:
        self._conn = sqlite3.connect(self.db_path)
        self._conn.row_factory = sqlite3.Row

        self._conn.execute("PRAGMA journal_mode = OFF")
        self._conn.execute("PRAGMA synchronous = OFF")
        self._conn.execute("PRAGMA temp_store = MEMORY")
        self._conn.execute("PRAGMA cache_size = -100000")
        self._conn.execute("PRAGMA locking_mode = EXCLUSIVE")
        self._conn.execute("PRAGMA mmap_size = 268435456")

    def __enter__(self) -> "Database":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def close(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None

    def connect(self) -> None:
        if not self._conn:
            self._setup_conn()

    def commit(self) -> None:
        self._conn.commit()

    def clone(self) -> "Database":
        return Database(self.db_path)

    @contextmanager
    def atomic(self):
        try:
            yield
            self._conn.commit()
        except Exception:
            self._conn.rollback()
            raise

    @classmethod
    def copy_to(cls, dir: str = None, filename: str = FILENAME) -> str:
        """Copy the database to current working directory, returning the new path."""
        path = Path(dir) if dir else Path.cwd()
        target_path = path / filename

        with resources.path(cls.PATH, cls.FILENAME) as db_file:
            shutil.copy(db_file, target_path)

        return str(target_path)

    def execute(self, query: str, params: Tuple | Dict = ()) -> sqlite3.Cursor:
        """Execute a single query"""
        cursor = self._conn.cursor()
        cursor.execute(query, params)
        return cursor

    def execute_many(
        self, query: str, params_list: List[Tuple | Dict]
    ) -> sqlite3.Cursor:
        """Execute query with multiple parameter sets"""
        cursor = self._conn.cursor()
        cursor.executemany(query, params_list)
        return cursor

    def insert_many(self, table: str, data_list: List[Dict[str, Any]]) -> None:
        """Insert multiple rows"""
        if not data_list:
            return

        column_list = list(data_list[0].keys())
        columns = ", ".join(column_list)
        placeholders = ", ".join(["?" for _ in column_list])
        query = f"INSERT INTO {table} ({columns}) VALUES ({placeholders})"

        params_list = [tuple(row[col] for col in column_list) for row in data_list]
        self.execute_many(query, params_list)

    def create_table(self, table_name: str, columns: str) -> None:
        """Create table with given columns definition"""
        query = f"CREATE TABLE IF NOT EXISTS {table_name} ({columns})"
        self.execute(query)
        self._conn.commit()

    def create_tables(self, tables: list[object]) -> None:
        for table in tables:
            if hasattr(table, "create_table") and callable(table.create_table):
                table.create_table()

    def drop_tables(self, tables: list[object]) -> None:
        for table in tables:
            if hasattr(table, "drop") and callable(table.drop):
                table.drop()

    def create_fts_table(
        self,
        table_name: str,
        columns: List[str],
    ) -> None:
        """Create FTS5 virtual table"""
        columns_str = ", ".join(columns)

        prefix_list = " ".join([str(i) for i in range(2, self.MAX_PREFIX + 1)])

        fts_options = [
            '''tokenize = "unicode61 remove_diacritics 2"''',
            f'''prefix="{prefix_list}"''',
        ]

        options_str = ", " + ", ".join(fts_options) if fts_options else ""

        query = f"""CREATE VIRTUAL TABLE IF NOT EXISTS "{table_name}" USING fts5({columns_str}{options_str})"""
        self.execute(query)
        self._conn.commit()

    def vacuum(self) -> None:
        """Vacuum database"""
        self.execute("VACUUM")

    def analyze(self) -> sqlite3.Cursor:
        """Analyze database for query optimization"""
        return self.execute("ANALYZE")

    def set_db_path(self, path: str) -> None:
        """Stores the database path in a config file for external storage and reloads the connection with the new path."""
        self.CONFIG_FILE.write_text(path)
        self.close()
        self.db_path = path
        self._setup_conn()

    @classmethod
    def get_db_path(cls) -> str:
        """Fetches the database path from config or defaults to the bundled .db file."""
        if cls.CONFIG_FILE.exists():
            return cls.CONFIG_FILE.read_text().strip()

        from importlib.resources import files

        db_file = files(cls.PATH) / cls.FILENAME
        return str(db_file)

    def revert_to_default(self):
        """Removes config file, external database and reverts to the bundled db file."""
        self.close()
        self.CONFIG_FILE.unlink()
        db_file = Path(self.db_path)
        db_file.unlink()
        self.db_path = self.get_db_path()
        self._setup_conn()


db = Database()


@atexit.register
def close_db() -> None:
    db.close()



from localis.search.search_engine import SearchEngine
from rapidfuzz import fuzz, process


class FuzzySearch(SearchEngine):
    NOISE_THRESHOLD = 0.6

    def score_candidate(self, candidate):
        score = 0.0
        total_matched_weight = 0.0

        for field, weight in self.field_weights.items():
            fvalue = getattr(candidate, field, "")
            if fvalue:
                values = fvalue.lower().split("|")
                matches = process.extract(
                    self.query,
                    values,
                    scorer=fuzz.token_set_ratio,
                    score_cutoff=60,
                )
                field_score = (
                    max(score for _, score, i in matches) / 100 if matches else 0.0
                )

                # # Filter noise
                if field_score >= self.NOISE_THRESHOLD:
                    score += field_score * weight
                    total_matched_weight += weight
        return score / total_matched_weight if total_matched_weight > 0 else 0.0



from localis.search.search_engine import SearchEngine
from localis.data import Model


class NgramSearch(SearchEngine):
    SCORE_THRESHOLD = 0.5
    NGRAM_LEN = 2

    def __init__(self, query, model_cls, field_weights, limit=None):
        # split query into tokens and ngram each one
        self.q_token_ngrams: list[set[str]] = [
            self._ngram(t, self.NGRAM_LEN) for t in query.split(" ")
        ]
        super().__init__(query, model_cls, field_weights, limit)

    def score_candidate(self, candidate: Model) -> float:
        score = 0.0

        f_token_map: dict[str, tuple[set[str], float]] = {}

        for field, weight in self.field_weights.items():
            value = getattr(candidate, field, "")
            for token in value.replace("|", " ").split():
                if token not in f_token_map:
                    f_token_map[token] = (self._ngram(token), weight)

        for q_token_ngram in self.q_token_ngrams:
            for f_token_ngrams, weight in f_token_map.values():
                if not q_token_ngram or not f_token_ngrams:
                    continue
                score += self._score_ngrams(q_token_ngram, f_token_ngrams) * weight

        return score

    def _score_ngrams(self, q_token_ngram: set[str], token_ngrams: set[str]) -> float:
        return len(q_token_ngram & token_ngrams) / len(q_token_ngram | token_ngrams)

    def _ngram(self, s: str, n: int = 2) -> set[str]:
        s = s.lower().replace(" ", "")
        return {s[i : i + n] for i in range(len(s) - n + 1)}



# Search Engine for registries.

from localis.data import Model
from localis.dtos import DTO
from abc import abstractmethod, ABC


class SearchEngine(ABC):
    MAX_ITERATIONS = 8
    MIN_TOKEN_LEN = 1
    MATCH_THRESHOLD = 0.6
    STRONG_MATCH_THRESHOLD = 0.85
    FTS_HARD_LIMIT = 2000

    def __init__(
        self,
        query: str,
        model_cls: type[Model],
        field_weights: dict[str, float],
        order_fields: list[str],
        limit: int = None,
    ):
        self.query: str = query.lower()
        self.tokens = self.query.split()
        self.model_cls: type[Model] = model_cls
        self.field_weights: dict[str, float] = field_weights
        self.order_fields: list[str] = order_fields
        self.limit: int | None = limit

        self._max_score = sum(field_weights.values())
        self._iterations = max(len(t) for t in self.tokens)

        self._scored: set[int] = set()
        self._matches: dict[Model, float] = {}
        self._iteration_match_counts = [0]
        self._has_strong_match = False

    def run(self) -> list[tuple[DTO, float]]:
        self.main()
        return self.results

    def main(self):
        candidates = self._fetch_candidates(exact=True)

        for i in range(1, self._iterations + 1):
            if self.query == "beuzchov":
                print(i, self.tokens)
                print(self._iterations)
            self._score_candidates(candidates)

            if self._should_exit_early(i):
                break

            self._truncate_tokens()

            candidates = self._fetch_candidates(i)

    def _fetch_candidates(self, i: int = None, exact=False) -> list[Model]:
        if i is None:
            limit = None
        else:
            if i == self._iterations - 1:
                limit = self.FTS_HARD_LIMIT
            else:
                limit = min(self.FTS_HARD_LIMIT, i * 400)

        fts_q = " ".join(self.tokens)

        order_by = ["rank"] if exact or i == self._iterations - 2 else []

        return self.model_cls.fts_match(
            fts_q, exact_match=exact, limit=self.FTS_HARD_LIMIT, order_by=order_by
        )

    def _score_candidates(self, candidates: list[Model]):
        for c in candidates:
            if c.id not in self._scored:
                score = self.score_candidate(c)
                if score >= self.MATCH_THRESHOLD:
                    self._matches[c] = score
                if score >= self.STRONG_MATCH_THRESHOLD:
                    self._has_strong_match = True
                self._scored.add(c.id)

    def _should_exit_early(self, i):
        if i == self._iterations:
            return True

        # Did we get a strong match?
        if self._has_strong_match:
            return True

        # Did the last two iteration produce a significant number of new matches?
        prev_idx = i - 1
        prev_match_count = self._iteration_match_counts[prev_idx]

        new_match_count = len(self._matches) - prev_match_count
        self._iteration_match_counts.append(new_match_count)

        if prev_match_count == 0:
            return False

        if i >= 3 and sum(self._iteration_match_counts[-2:]) <= 1:
            return True

        return False

    @abstractmethod
    def score_candidate(self, candidate: Model) -> float:
        return 0.0

    def _truncate_tokens(self) -> None:
        self.tokens = [
            t[:-1] if len(t) > self.MIN_TOKEN_LEN else t for t in self.tokens
        ]

    def _at_limit(self) -> bool:
        if self.limit is not None:
            return len(self._matches) >= self.limit
        return False

    @property
    def results(self) -> list[tuple[DTO, float]]:
        return sorted(
            [(m.to_dto(), score) for m, score in self._matches.items()],
            key=lambda x: (x[1], *[getattr(x[0], f) for f in self.order_fields]),
            reverse=True,
        )[: self.limit]



from localis.data.models.model import Model
from localis.data.models.country_model import CountryModel
from localis.data.models.subdivision_model import SubdivisionModel
from localis.data.models.city_model import CityModel



# These DTOS are the final product delivered to the user.

from dataclasses import dataclass, asdict
import json
from abc import ABC


@dataclass(slots=True)
class DTO(ABC):
    id: int
    name: str
    alt_names: list[str]

    def to_dict(self):
        return asdict(self)

    def json(self):
        return json.dumps(self.to_dict(), indent=2)

    def __str__(self):
        return self.json()


@dataclass(slots=True)
class Country(DTO):
    id: int
    name: str
    official_name: str
    alpha2: str
    alpha3: str
    numeric: int
    alt_names: list[str]
    flag: str


@dataclass(slots=True)
class SubdivisionBasic:
    name: str
    geonames_code: str
    iso_code: str
    admin_level: int


@dataclass(slots=True)
class Subdivision(DTO):
    id: int
    name: str
    alt_names: list[str]
    type: str
    geonames_code: str
    iso_code: str
    admin_level: int
    country: str
    country_alpha2: str
    country_alpha3: str
    parent_id: int | None


@dataclass(slots=True)
class City(DTO):
    id: int
    geonames_id: int
    name: str
    display_name: str | None
    subdivisions: list[SubdivisionBasic]
    country: str
    country_alpha2: str
    country_alpha3: str
    alt_names: list[str]
    population: int | None
    lat: float
    lng: float



from localis.data import db, Database
from localis.data.models.fields import Field, Expression
from localis.dtos import DTO
from typing import TypeVar, Generic
from abc import ABC
import sqlite3
from typing import Type
import json
from abc import abstractmethod
from localis.utils import prep_tokens

TDTO = TypeVar("TDTO", bound=DTO)


class Model(Generic[TDTO], ABC):
    db: Database = db
    table_name = ""
    dto_class: Type[TDTO] = None

    id: int
    rank: float

    def __init__(self, id: int, rank: float | None = None, **kwargs):
        self.id = id
        self.rank = rank

    @classmethod
    def from_row(cls, row: sqlite3.Row):
        return cls(**dict(row))

    @abstractmethod
    def to_dto(self) -> TDTO:
        data = self.__dict__
        data.pop("rank")
        return self.dto_class(**data)

    @classmethod
    def create_table(cls) -> None:
        cls._create_fts()
        cls.db.commit()

    @classmethod
    def count(cls) -> int:
        return cls.db.execute(f"SELECT COUNT(*) FROM {cls.table_name}").fetchone()[0]

    @classmethod
    def columns(cls):
        for name, attr in cls.__dict__.items():
            if isinstance(attr, Field):
                if not attr.index:
                    name = f"{name} UNINDEXED"
                yield name

    @classmethod
    def _create_fts(cls) -> None:
        cls.db.create_fts_table(cls.table_name, cls.columns())

    @classmethod
    def drop(cls):
        cls.db.execute(f"DROP TABLE IF EXISTS {cls.table_name}")
        cls.db.commit()
        cls.db.vacuum()

    @classmethod
    def insert_many(cls, data: list[dict]) -> None:
        """Insert multiple rows into the database, requires with atomic."""
        cls.db.insert_many(cls.table_name, data)

    def __str__(self):
        return json.dumps(self.__dict__, indent=4)

    @classmethod
    def get(cls, expr: Expression):
        row = cls.db.execute(
            f"SELECT rowid as id, * FROM {cls.table_name} WHERE {expr.sql} LIMIT 1",
            expr.params,
        ).fetchone()
        if not row:
            return None
        return cls.from_row(row)

    @classmethod
    def get_by_id(cls, id: int):
        row = cls.db.execute(
            f"SELECT rowid as id, * FROM {cls.table_name} WHERE id = ? LIMIT 1", (id,)
        ).fetchone()
        if not row:
            return None
        return cls.from_row(row)

    @classmethod
    def select(
        cls,
        expr: Expression | None = None,
        order_by: str | None = None,
        limit: int | None = None,
    ) -> list["Model"]:

        order_by = f"ORDER BY {order_by}" if order_by else ""
        limit = f"LIMIT {limit}" if limit else ""

        if expr:
            rows = cls.db.execute(
                f"SELECT rowid as id, * FROM {cls.table_name} WHERE {expr.sql} {order_by} {limit}",
                expr.params,
            ).fetchall()
        else:
            rows = cls.db.execute(
                f"SELECT rowid as id, * FROM {cls.table_name} {order_by} {limit}"
            ).fetchall()
        return [cls.from_row(row) for row in rows]

    @classmethod
    def fts_match(
        cls,
        query: str = None,
        field_queries: dict[str, str] = None,
        exact_match: bool = True,
        order_by: list[str] = [],
        limit: int = None,
        offset: int = None,
    ):
        if field_queries:
            params = list(prep_tokens(q, exact_match) for q in field_queries.values())
            q_where = "WHERE " + "AND ".join(
                f"{col} MATCH ?" for col in field_queries.keys()
            )
        elif query:
            sanitized_input = prep_tokens(query, exact_match)

            params = [sanitized_input]
            q_where = f"WHERE {cls.table_name} MATCH ?"
        else:
            return []

        bm25 = f", bm25({cls.table_name}, 50.0) as rank" if "rank" in order_by else ""

        q_order_by = "ORDER BY " + ", ".join(order_by) if order_by else ""

        q_limit = "LIMIT ?" if limit is not None else ""

        q_offset = f"OFFSET ?" if offset is not None else ""

        q = f"""SELECT rowid as id, *{bm25} FROM {cls.table_name}
                    {q_where}
                    {q_order_by}
                    {q_limit}
                    {q_offset}"""

        if limit:
            params.append(limit)

        if offset:
            params.append(offset)

        cursor = cls.db.execute(q, params)
        rows: list[sqlite3.Row] = cursor.fetchall()
        cursor.close()
        return [cls.from_row(row) for row in rows if row]



from localis.registries.registry import Registry
from localis.data import CountryModel, Country


class CountryRegistry(Registry[CountryModel, Country]):
    """
    Registry for managing Country entities.

    Supports get by alpha2, alpha3 & numeric codes, lookup by country name with support for some aliases/former names,
    and fuzzy search.
    """

    ID_FIELDS = ("id", "alpha2", "alpha3", "numeric")

    SEARCH_FIELD_WEIGHTS = {
        "name": 1.0,
        "official_name": 1.0,
        "alt_names": 1.0,
    }

    def get(
        self,
        *,
        id: int = None,
        alpha2: str = None,
        alpha3: str = None,
        numeric: int = None,
        **kwargs,
    ):
        cls = self._model_cls

        field_map = {
            "id": None,
            "alpha2": cls.alpha2,
            "alpha3": cls.alpha3,
            "numeric": cls.numeric,
        }

        model = None
        for arg, field in field_map.items():
            val = locals()[arg]
            if val is not None:
                model = cls.get_by_id(val) if arg == "id" else cls.get(field == val)

        return model.to_dto() if model is not None else None

    def filter(
        self,
        query=None,
        name=None,
        limit=None,
        official_name: str = None,
        alt_name: str = None,
        **kwargs,
    ):
        if kwargs:
            return []
        if official_name:
            kwargs["official_name"] = official_name
        if alt_name:
            kwargs["alt_names"] = alt_name

        return super().filter(query, name, limit, **kwargs)

    @property
    def _sql_filter_base(self):
        return f"""SELECT c.*, f.tokens FROM countries_fts f
        JOIN countries c ON f.rowid = c.id
        """
